/*
	tgf-gostream.cc, part of
	
	treegraph
	
	Tree formatting program
	
	Generates vector graphics (SVG,EPS) from .tgf-tree description files.

	Copyright (c) 2003-06 by Joern Mueller


	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program (GPL.html); if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
	(also: http://www.gnu.org)
*/







#include "tgf-gostream.h"
#include "tgf-tree.h"
#include "tgf-main.h"
#include <cstdio>

/*
**
**
**
** gostream
**
**
**
*/


gostream::gostream(std::ostream &s1) : s(s1)
{
	fontpath="fonts.txt";
}


void gostream::set_fontpath(std::string& path)
{
	fontpath=path;
}


/*
**
**
**
** svg_ostream
**
**
**
*/

const float PiOver2=asin(1.0);

svg_ostream::svg_ostream(std::ostream& s1,const tree_desc& dsc) : gostream(s1), dsc1(dsc)
{

}

void svg_ostream::moveto(float x,float y)
{
	s << 'M'<< x << ' ' << y;

}


void svg_ostream::lineto(float x,float y)
{
	s << 'L' << x << ' ' << y;

}

void svg_ostream::curveto(float x1,float y1, float x2,float y2, float x3,float y3)
{
	s<< 'C' << x1 <<' ' << y1 << ' '<< x2 <<' ' << y2 << ' '<< x3 <<' ' << y3; 
}
			
void svg_ostream::write_head(const fRect& paper, const fRect& bbox, float lthick)
{
	s << "<?xml version=\"1.0\" standalone=\"no\"?>\n";
	s << "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 20010904//EN\"\n";
	s << "\"http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd\">\n";
	s << "<svg ";
//	s << "width=\""<<paper.right<<"px\"\n";
//	s << "height=\""<<paper.bottom<<"px\"\n";
	s << "viewBox=\""<<bbox.left<< " " << bbox.top << " "<<bbox.right<< " " << bbox.bottom << "\"\n";
	s << "xmlns=\"http://www.w3.org/2000/svg\"\n";
	s << "xmlns:xlink=\"http://www.w3.org/1999/xlink\"\n";
	s << "width=\""<<ceil(bbox.right-bbox.left)<<"px\" height=\""<<ceil(bbox.bottom-bbox.top)<< "px\"\n";
	s << "transform=\"translate("<<0.5*(paper.right-bbox.right) <<","<< 0.5*(paper.bottom-bbox.bottom)<< ")\"";
 	s << ">\n";

	s<< "<defs>\n";
	s<< "<style type=\"text/css\"><![CDATA[\n";
	s<< "path {\n";
	s<< "fill:none;\n";
	s<< "stroke:#000000;\n";
	s<< "stroke-linecap:round;\n";
	s<< "stroke-width:0.850394;\n";
	s<< "fill-opacity:1; \n";
	s<< "fill-rule:evenodd; \n";
	s<< "stroke-opacity:1;\n";
	s<< "}\n";
	s<< "line {\n";
	s<< "fill:none;\n";
	s<< "stroke:#000000;\n";
	s<< "stroke-linecap:round;\n";
	s<< "stroke-width:0.850394;\n";
	s<< "fill-opacity:1; \n";
	s<< "fill-rule:evenodd; \n";
	s<< "stroke-opacity:1;\n";
	s<< "}\n";

//Verschiedene Text-styles
//Default-font
	
	fontstyle f1=dsc1.getstyle("default");
	s<< "text {\n";
	s<< "fill: #000000;\n";
	s<< "font-family:Sans-Serif; \n";
	if(f1.face==fs_italic){
		s << "font-weight:normal;\nfont-style:italic;\n";
	} else if(f1.face==fs_bold){
		s << "font-style:normal;\nfont-weight:bold;\n";
	} else {
		s<< "font-style:normal;\nfont-weight:normal;\n";
	}
	s<<"font-size:"<<f1.fontsize<<"px;\n";	
	s<< "}\n";
	

	
	for(std::list<namedfontstyle>::const_iterator fit=dsc1.style.begin();fit!=dsc1.style.end();++fit){
		if(fit->name=="default"){
			continue;
		} else {
				s<< "text."<<fit->name<<" {\n";
				if(fit->face==fs_italic){
					s << "font-weight:normal;\nfont-style:italic;\n";
				} else if(fit->face==fs_bold){
					s << "font-style:normal;\nfont-weight:bold;\n";
				} else {
					s<< "font-style:normal;\nfont-weight:normal;\n";
				}
				s<<"font-size:"<<fit->fontsize<<"px;\n";	
				s<<"}\n";
		}
	}

//label-beschr. Ueberschreibt obien ndl-Eintrag (wegen Farbe rot)
	s<< "text.ndl {\n";
	s<< "fill: #FF0000;\n";
	s<< "font-size: 10px;\n";
	s<< "font-weight:bold;\n";
	s<< "}\n";

	s<< "]]>\n";
	s<< "</style>\n";
	s<< "</defs>\n";

 
}

void svg_ostream::write_foot()
{
	s << "\n</svg>\n";
	s.flush();
}

void svg_ostream::write_text(const nfstring& str, const fPoint& p, float angle)
{
	//class nur schreiben, wenn wirklich nicht default
	if(str.stylename=="default" || !dsc1.styleexists(str.stylename)){
		s<< "<text ";
	} else {
		s<< "<text class=\""<<str.stylename<<"\" ";
	}
	if(angle!=0.0){
		s << "x=\"0\" y=\"0\"";
		s << "\ntransform=\"matrix(";
		float c1=cos(angle),s1=sin(angle);
		s<<c1<<","<< s1<<","<<-s1<<","<<c1<<",";
		s<<p.x<<","<<p.y<<")\"";
	} else {
		s << " x=\""<<p.x<<"\" y=\""<<p.y<<"\"";
	}
	s <<">";	//Klammern aendern
	std::string sneu="";
	for(unsigned int i=0;i<str.length();i++){
		if(str[i]=='<') sneu+="&lt;";
		else if(str[i]=='>') sneu+="&gt;";
		else sneu+=str[i];
	}
	s << sneu;
	s << "</text>\n";
	
}

void svg_ostream::write_rgabel(float l1,float y1 ,float l2,float y2,float xx,float maxr)
{
	float hdmt=0.5*maxr;

	s << "<path d=\"";
	s << "M" << xx+l1 << " " << y1;
	s << "H" << xx+maxr;
	s << "C" << xx+hdmt << " " << y1 << " " << xx << " " << y1+hdmt;
	s << " " << xx << " " << y1+maxr;
	s << "V" << y2-maxr;
	s << "C" << xx << " " << y2-hdmt << " " << xx+hdmt << " " << y2;
	s << " " << xx+maxr << " " << y2;
	s << "H" << xx+l2;
	s << "\" />\n";//path

}

void svg_ostream::write_egabel(float l1,float y1,float l2,float y2,float xx)
{
	s << "<path d=\"";
	s << "M" << xx+l1 << " " << y1;
	s << "H" << xx;
	s << "V" << y2;
	s << "H" << xx+l2;
	s << "\" />\n";//path
}

void svg_ostream::write_hline(float x1,float y1,float l)
{
	s << "<line x1=\"" << x1 << "\" y1=\"" << y1 << "\" ";
	s << "x2=\"" << x1+l << "\" y2=\"" << y1 << "\" />\n";

}

void svg_ostream::write_klammer(const nfstring& txt,float y1,
	float y2,float x,float ext, int style)
{

	y1-=ext;
	y2+=ext;
	
	if(!(style&klammer::eckig)){

	s << "<path style=\"stroke-linejoin:round;stroke-linecap:round;stroke-width:0.8\"\n";
	
	float ll=0.5*fabs(y2-y1);
	if(ll>20.0) ll=20.0;
	float frc=0.035*ll;
	
	float md=0.5*(y1+y2);
	s << "d=\"";

	moveto(x,y2);
	curveto(x+3,y2-4*frc, x+3,y2-4*frc, x+3,y2-8*frc);
	lineto(x+3,md+7*frc);
	curveto(x+3,md+5*frc, x+3,md+2*frc, x+6,md);

	moveto(x+6,md);
	curveto(x+3.5,md+3*frc, x+3.5,md+5*frc, x+3.5,md+7*frc);
	lineto(x+3.5,y2-8*frc);
	curveto(x+3.5,y2-5*frc, x+3.5,y2-3*frc, x,y2);

	moveto(x,y1);
	curveto(x+3,y1+4*frc, x+3,y1+4*frc, x+3,y1+8*frc);
	lineto(x+3,md-7*frc);
	curveto(x+3,md-5*frc, x+3,md-2*frc, x+6,md);

	moveto(x+6,md);
	curveto(x+3.5,md-3*frc, x+3.5,md-5*frc, x+3.5,md-7*frc);
	lineto(x+3.5,y1+8*frc);
	curveto(x+3.5,y1+5*frc, x+3.5,y1+3*frc, x,y1);
	s<<"\"\n/>";

	} else {
	s << "<g>";
	s << "<path style=\"stroke-linejoin:round;stroke-linecap:butt;stroke-width:0.8\" ";
	s << "d=\"M"<<x<<' '<<y1;
	s << " H"<<x+4.0<<"V"<<y2<<"H"<<x<<"\" />\n";
	s << "<path style=\"stroke-linecap:butt;stroke-width:1\" ";
	s << "d=\"M"<<x+4.0-0.7<<' '<<y1<<"V"<<y2<<"\" />\n";
	s << "</g>";
	}




	fPoint m;
	if(style&klammer::gedreht){
		m.x=x+8.0;
		m.y=(y1+y2)/2+0.33*dsc1.getfontsize(txt);
	} else {
		m.x=x+8.0+dsc1.getfontsize(txt);
		m.y=(y1+y2)/2+0.5*dsc1.helv_stringwidth(txt);
	}
	write_text(txt,m,(style&klammer::gedreht)?0:-PiOver2);


}



void svg_ostream::write_rule(const fPoint& pos, float z,float wd)
{
	s<<"<g style=\"stroke-linecap:butt;stroke-width:1\">\n";
	write_hline(pos.x,pos.y,wd);
	s<<"<g style=\"stroke-width:0.3\">\n";

	
	float x1,l;
	for(int i=0; i<=10; i++){
		x1=pos.x+i*0.1*wd;
		l=2.0;
		if(i==0 || i==10)
			l=3.0;
		else if(i==5)
			l=2.5;
		s << "<line x1=\"" << x1 << "\" y1=\"" << pos.y-l << "\" ";
		s << "x2=\"" << x1 << "\" y2=\"" << pos.y+l << "\" />\n";
	}
	s<< "</g></g>\n";


	char st[10];
	sprintf(st,"%.5g",z);
	nfstring ruletext=nfstring(st,"rule");
	fPoint m;
	m.x=pos.x+wd/2-0.5*dsc1.helv_stringwidth(ruletext);
	m.y=pos.y+16;
	write_text(ruletext,m);

}

void svg_ostream::write_arrow(const fPoint& pos, float angle_deg)
{
std::cerr << "Arrow is not yet supported\n";
	const float htief=9,hhbr=3.4;
	s<<"<g transform=\"translate("<<pos.x<<","<<pos.y<<")\">\n";
	s<<"<g transform=\"rotate("<<angle_deg<<")\">\n";
	s << "<path style=\"stroke-linejoin:round;stroke-width:0.8\" ";
	s << "d=\"M0 0L"<<htief<<" "<<hhbr;
//	s << " H"<<x+4.0<<"V"<<y2+ext<<"H"<<x<<"\" />\n";
	
	s<<"</g></g>";
}

void svg_ostream::write_nodelabel(const fPoint& pos, const std::string& str)
{
	fontstyle tmp;
	tmp.fontsize=10;
	tmp.face=fs_bold;
	
	s << "<text class=\"ndl\" ";
	s << "x=\""<<pos.x<<"\" y=\""<<pos.y<<"\">";
	s << str;
	s << "</text>\n";
		
}



/*
**
**
**
** ps_ostream
**
**
**
*/
ps_ostream::ps_ostream(std::ostream& s1, const tree_desc& dsc) : gostream(s1), dsc1(dsc)
{
	finited=false;
}

void ps_ostream::setfont(const fontstyle& f)
//Reduces number of calls to scalefont/Makefont
{
//neu: verwendet Nimbus-Fonts

	bool ch=(!finited||(f.face!=laststyle.face)||(f.fontsize!=laststyle.fontsize));
	if(ch){
		const float& q=f.fontsize;
		if(f.face==fs_italic){
			s <<"/NimbusSanL-Regu findfont ["<<q<<" 0 "<<0.25*q<<" "<<q<< " 0 0] makefont setfont\n";
		} else if(f.face==fs_bold) {
			s <<"/NimbusSanL-Bold findfont "<<q<<" scalefont setfont\n";
		} else {
			s <<"/NimbusSanL-Regu findfont "<<q<<" scalefont setfont\n";
		}
		laststyle=f;
		finited=true;
	}
}


void ps_ostream::moveto(float x,float y)
{
	s << x << ' ' << ty(y) << " M ";

}


void ps_ostream::lineto(float x,float y)
{
	s << x << ' ' << ty(y) << " L ";

}

void ps_ostream::curveto(float x1,float y1, float x2,float y2, float x3,float y3)
{
	s<< x1 <<' ' << ty(y1) << ' '<< x2 <<' ' << ty(y2) << ' '<< x3 <<' ' << ty(y3) << " C "; 
}


		
void ps_ostream::write_head(const fRect& paper, const fRect& bbox, float lthick)
{

//bbox auf paper zentrieren.
	yh=bbox.bottom; //fuer rhd.-Koordinaten unter PostScript
	fPoint st(0.5*(paper.right-bbox.right),0.5*(paper.bottom-bbox.bottom));
	s << "%!PS-Adobe-3.0 EPSF-3.0\n";
	s << "%%Creator: treegraph " TGFVERSION "\n";
	s << "%%BoundingBox: " << int(st.x-lthick/2) << ' ' << int(st.y) << ' ';
	s << int(st.x+bbox.right+lthick+0.5) << ' ' << int(st.y+bbox.bottom+0.5) <<std::endl;
//		s << "%%DocumentNeededFonts: Helvetica Helvetica-Bold\n";//Italic wird aus Plain konstruiert
	s << "%%EndComments\n";
	s << "%%BeginProlog\n";
//	s << "150 dict begin "; 
	s << "/M {moveto} bind def /L {lineto} bind def /C {curveto} bind def\n";
	s << "/N {newpath} bind def /SW {show} bind def /S {stroke} bind def\n";
	s << "/LS {lineto stroke} bind def\n";
	s << "/SF {setfont} bind def /GS {gsave} bind def /GR {grestore} bind def\n";
	s << st.x <<' '<< st.y<< " translate\n";
	s << lthick << " setlinewidth 1 setlinecap\n";
 	s << "\n%%EndProlog\n";

	//hier werden die Nimbus-Fonts geladen:
	//NimbusSanL-Bold
	//NimbusSanL-Regu
	std::ifstream fonts(fontpath.c_str());
	if(!fonts){
		std::cerr << "Error: Could not find file \""<<fontpath<<"\"\n";
		return;
	}
	
	int ch;
	while((ch=fonts.get())!=EOF){
		s.put(ch);
	}
	fonts.close();
	
}

void ps_ostream::write_foot()
{
	s << "\n%%EOF\n";
	s.flush();
//Hier noch was wegen dict
}

void ps_ostream::write_text(const nfstring& str, const fPoint& p, float angle)
{
	setfont(dsc1.getstyle(str.stylename));
	s << p.x << " " << ty(p.y) << " M\n";
	
	if(angle!=0.0){
		s << "GS " << -angle*90/PiOver2 << " rotate ";
	}
	//Klammern aendern
	std::string sneu="";
	for(unsigned int i=0;i<str.length();i++){
		if(str[i]=='(') sneu+="\\(";
		else if(str[i]==')') sneu+="\\)";
		else sneu+=str[i];
	}
	
	s << "("<<sneu<<") SW\n";
	if(angle!=0.0){
		s << "GR\n";
	}

}

void ps_ostream::write_rgabel(float l1,float y1 ,float l2,float y2,float xx,float maxr)
{
	float hdmt=0.5*maxr;
//	s << "N ";
	moveto(xx+l1,y1);
	lineto(xx+maxr,y1);
	curveto(xx+hdmt,y1,xx,y1+hdmt,xx,y1+maxr);
	s<<std::endl;
	lineto(xx,y2-maxr);
	curveto(xx,y2-hdmt,xx+hdmt,y2,xx+maxr,y2);
	lineto(xx+l2,y2);
	s << "S\n";
	
	
}

void ps_ostream::write_egabel(float l1,float y1,float l2,float y2,float xx)
{
//	s << "N ";
	moveto(xx+l1,y1);
	lineto(xx,y1);
	lineto(xx,y2);
	lineto(xx+l2,y2);
	s << "S\n";
}

void ps_ostream::write_hline(float x1,float y1,float l)
{
//	s << "N ";
	moveto(x1,y1);
	lineto(x1+l,y1);
	s << "S\n"; 
}

void ps_ostream::write_klammer(const nfstring& txt,float y1,
	float y2,float x,float ext, int style)
{
	
	y1-=ext;
	y2+=ext;
	
	s << "GS\n";
	
	if(!(style&klammer::eckig)){
	
	float ll=0.5*fabs(y2-y1);
	if(ll>20.0) ll=20.0;
	float frc=0.035*ll;
	
	float md=0.5*(y1+y2);
	s << "0.8 setlinewidth 2 setlinejoin 1 setlinecap N\n";

	moveto(x,y2);
	curveto(x+3,y2-4*frc, x+3,y2-4*frc, x+3,y2-8*frc);
	lineto(x+3,md+7*frc);
	curveto(x+3,md+5*frc, x+3,md+2*frc, x+6,md);
	s << std::endl;

	moveto(x+6,md);
	curveto(x+3.5,md+3*frc, x+3.5,md+5*frc, x+3.5,md+7*frc);
	lineto(x+3.5,y2-8*frc);
	curveto(x+3.5,y2-5*frc, x+3.5,y2-3*frc, x,y2);
	s<<std::endl;

	moveto(x,y1);
	curveto(x+3,y1+4*frc, x+3,y1+4*frc, x+3,y1+8*frc);
	lineto(x+3,md-7*frc);
	curveto(x+3,md-5*frc, x+3,md-2*frc, x+6,md);
	s << std::endl;

	moveto(x+6,md);
	curveto(x+3.5,md-3*frc, x+3.5,md-5*frc, x+3.5,md-7*frc);
	lineto(x+3.5,y1+8*frc);
	curveto(x+3.5,y1+5*frc, x+3.5,y1+3*frc, x,y1);

	} else {
	
	s << "0.8 setlinewidth 2 setlinejoin 0 setlinecap N\n";
	moveto(x,y1);
	lineto(x+4.0,y1);
	lineto(x+4.0,y2);
	lineto(x,y2);
	s << "S\n";
	s << "1 setlinewidth\n";
	moveto(x+4.0-0.7,y1);
	lineto(x+4.0-0.7,y2);
	
	}
	
	
	s << "S GR\n";
	
	fPoint m;
	if(style&klammer::gedreht){
		m.x=x+8.0;
		m.y=(y1+y2)/2+0.33*dsc1.getfontsize(txt);
	} else {
		m.x=x+8.0+dsc1.getfontsize(txt);
		m.y=(y1+y2)/2+0.5*dsc1.helv_stringwidth(txt);
	}
	write_text(txt,m,(style&klammer::gedreht)?0:-PiOver2);

}



void ps_ostream::write_rule(const fPoint& pos, float z,float wd)
{

	s<<"GS 0 setlinecap 1 setlinewidth\n";
	write_hline(pos.x,pos.y,wd);
	s<<"0.3 setlinewidth\n";
	
	float x1,l;
	for(int i=0; i<=10; i++){
		x1=pos.x+i*0.1*wd;
		l=2.0;
		if(i==0 || i==10)
			l=3.0;
		else if(i==5)
			l=2.5;
	//	s << "N ";
		moveto(x1,pos.y-l);
		lineto(x1,pos.y+l);
		s<<"S\n";
	}
	s<< "GR\n";

	
	char st[10];
	sprintf(st,"%.5g",z);
	nfstring ruletext=nfstring(st,"rule");
	fPoint m;
	m.x=pos.x+wd/2-0.5*dsc1.helv_stringwidth(ruletext);
	m.y=pos.y+16;
	write_text(ruletext,m);

}

void ps_ostream::write_arrow(const fPoint& pos, float angle_deg)
{
/*
	const float htief=9,hhbr=3.4,kerb=2.4;
	s<<"<g transform=\"translate("<<pos.x<<","<<pos.y<<")\">\n";
	s<<"<g transform=\"rotate("<<angle_deg<<")\">\n";
	s << "<path style=\"stroke-linejoin:round;stroke-width:0.8\"";
	s << "d=\"M0 0L"<<htief<<" "<<hhbr;
//	s << " H"<<x+4.0<<"V"<<y2+ext<<"H"<<x<<"\" />\n";
	
	s<<"</g></g>";
*/
}

void ps_ostream::write_nodelabel(const fPoint& pos, const std::string& str)
{
	s<<"1 0 0 setrgbcolor ";
	write_text(nfstring(str,"ndl"),pos);	
	s << "0 0 0 setrgbcolor \n";
		
}



