/*
	tgf-parser.cc, part of
	
	treegraph
	
	Tree formatting program
	
	Generates vector graphics (SVG,EPS) from .tgf-tree description files.

	Copyright (c) 2003-06 by Joern Mueller


	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program (GPL.html); if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
	(also: http://www.gnu.org)
*/

#include "tgf-parser.h"
#include "tgf-error.h"
#include <cstdio>

/*
line endings:
Unix: 0x0A
Mac: 0x0D
Win: 0x0D 0x0A

Mac compilers treat '\n' as 0x0D
Problem: Mac file on Unix|Win System, Unix file on Mac system

*/

parser::parser(std::ifstream& s1) : ein(s1), line(1), lastRead(EOF)
{

}

int parser::lies()
//nur hier zwischen Endungen Unterscheiden. Rest funktioniert mit \n
{
	int ch=ein.get();
	if(ch==0x0D){
		++line;
		if(ein.peek()==0x0A){//DOS
			ein.get(); 
		}
		ch='\n';
	} else if(ch==0x0A){
		++line;
		ch='\n';
	}
	lastRead=ch;
	return ch;
}

void parser::zurueck()
//kann nur einmal aufgerufen werden!
{
	ein.unget();
	if(lastRead=='\n')
		--line;
}

void parser::skipspaces()
{
	int ch;
	while((ch=lies())!=EOF) {
		if(!isspace(ch)){
			zurueck();
			return;
		}
	}
}


bool parser::eof() const {
	return parser::ein.eof();
}
bool parser::good() const
{
	return parser::ein.good();
}




void parser::fehler(const std::string& s)
{
	throw tgfError::syntax(s,line);
}

bool parser::lcEqual(const std::string& a, const std::string& bLower)
// a==bLower, where bLower is a lower case std::string
{
	if(a.size()!=bLower.size()) return false;
	std::string::const_iterator pa=a.begin();
	std::string::const_iterator pb=bLower.begin();
	while(pb!=bLower.end()){
		if(tolower(*pa)!= *pb) return false;
		++pa; 
		++pb;
	}
	return true;
}


tgfParser::tgfParser(std::ifstream& s1) : parser(s1)
{

}


float tgfParser::readbracefloat()
// liest {<float>}
{
	std::string s;
	readbracestring(s);
	float f;
	if(std::sscanf(s.c_str(),"%f",&f)!=1){
		fehler("expected {float}");
	}

	return f;
	
}

bool tgfParser::skipbracestring()
//{<string>} ignorieren, wenn vorhanden
// true, wenn Klammer vorhanden war
{
	skipspaces();
	int ch=tgfParser::lies();
	if(ch!='{'){
		tgfParser::zurueck();
		return false;
	}
	int bb=1;
	while((ch=tgfParser::lies())!=EOF){
		if(ch=='}') --bb;
		else if(ch=='{') ++bb;
		else if(ch=='\n'){
			fehler("Line break in argument");
		}
		if(bb==0) break;
	}
	if(bb>0){
		fehler("Missing }");
	}
	return true;
}


void tgfParser::readbracestring(std::string& str)
//{<string>}
{
	skipspaces();
	int ch=tgfParser::lies();
	if(ch!='{'){
		fehler("Missing {");
	}
	str="";
	int bb=1;
	while((ch=tgfParser::lies())!=EOF){
		if(ch=='}') --bb;
		else if(ch=='{') ++bb;
		else if(ch=='\n'){
			fehler("Line break in argument");
		}
		if(bb==0) break;
		else if(ch=='\t') ch=' ';
		str+=ch;
	}
	if(bb>0){
		fehler("Missing }");
	}
}



void tgfParser::readcommand(std::string& cmd)
//beginnt hinter '\' zu lesen
{
	const int mwlen=20;
	char wort[mwlen+1];
	int i;
	for(i=0; i<mwlen; i++){
		int ch=tgfParser::lies();
		if(isalnum(ch) || ch=='_' || ch=='*')
			wort[i]=ch;
		else {
			parser::zurueck();
			break;
		}
	}
	wort[i]=0;
	cmd=wort;
}

void tgfParser::skipcomment()
//beginnt hinter '%' zu lesen
{
	int ch;
	while((ch=tgfParser::lies())!=EOF){
		if(ch=='\n') return;
	}
	
}




tgfParser::ttoken tgfParser::readnextword(std::string& s, bool descriptionpart)
{
	s="";
	ttoken was=kErr;
	int ch;
	do {
		if(descriptionpart){
		//auch newlines uebergeben
			int line1=curline();
			skipspaces();
			int line2=curline();
			for(int k=line1;k<line2;k++){
				s+='\n';
			}
			if(line2>line1) return kNewline;
			
		} else {
			skipspaces();
		}
		ch=tgfParser::lies();
		switch(ch){
		case '(': was=kOpenTree; break;
		case ')': was=kCloseTree; break;
		case ',': was=kNextTree; break;
		case EOF: was=kEOF; break;
		case '\\':
			readcommand(s);
			if(s=="begindef"){
				was=kBeginDef;
			} else if(s=="enddef"){
				was=kEndDef;
			} else {
				was=kCommand;
			}
			break;
		case '"':
			while((ch=tgfParser::lies())!=EOF){
				if(ch=='"') break;
				else s+=(char)ch;
			}
			was=kText;
			break;
		case '%':
			if(descriptionpart){
				while((ch=tgfParser::lies())!=EOF){
					if(ch=='\n'){
						tgfParser::zurueck();
					//newline selbst nicht uebergeben
						s="%"+s;
						return kComment;
					} else s+=(char)ch;
				}
			} else {
				skipcomment();
			}
			break;
		default: 
			fehler("readnextword: Illegal character '"+tgfError::CharToString(ch)
				 +"' (ASCII " +tgfError::IntegerToString(ch)+")");
			break;
		}
	} while (ch=='%');
	return was;

}






