/*
	tgf-render.cc, part of
	
	treegraph
	
	Tree formatting program
	
	Generates vector graphics (SVG,EPS) from .tgf-tree description files.

	Copyright (c) 2003-06 by Joern Mueller


	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program (GPL.html); if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
	(also: http://www.gnu.org)
*/







#include "tgf-main.h"
#include "tgf-tree.h"
#include "tgf-render.h"
#include "tgf-utils.h"
#include "tgf-error.h"
#include <cmath>


void writeInnerNodeTextLine(const fPoint& p,const nfstring& s1, 
	const nfstring& s2,  const nfstring& sepw, 
const tree_desc& dsc , gostream *of);

	
void tree::render(gostream *of) const
{
	fRect paper,bbox;
	paper.right=595;
	paper.bottom=842;

	// Verschiedene Papierformate 
	if(dsc.paper=="a4"){
		paper.right=595; paper.bottom=842;
	} else if(dsc.paper=="a3"){
		paper.right=842; paper.bottom=1191;
	} else if(dsc.paper=="a5"){
		paper.right=420; paper.bottom=595;
	} else if(dsc.paper=="a0"){
		paper.right=2384; paper.bottom=3370;
	} else if(dsc.paper=="a1"){
		paper.right=1684; paper.bottom=2384;
	} else if(dsc.paper=="a2"){
		paper.right=1191; paper.bottom=1684;
	} else if(dsc.paper=="a6"){
		paper.right=297; paper.bottom=420;
	} else if(dsc.paper=="a7"){
		paper.right=210; paper.bottom=297;
	} else if(dsc.paper=="a8"){
		paper.right=148; paper.bottom=210;
	} else if(dsc.paper=="a9"){
		paper.right=105; paper.bottom=148;
	} else if(dsc.paper=="11x17"){
		paper.right=792; paper.bottom=1224;
	} else if(dsc.paper=="legal"){
		paper.right=612; paper.bottom=1008;
	} else if(dsc.paper=="letter"){
		paper.right=612; paper.bottom=792;
	} else if(dsc.paper=="bbox"){
		paper.right=dsc.picwidth;
		paper.bottom=dsc.picheight;
	} else {
		throw tgfError::error("Unknown paper format '"+dsc.paper+"'");
	}
	if(dsc.landscape){
		float tmp=paper.right;
		paper.right=paper.bottom;
		paper.bottom=tmp;
	}

	bbox.right=dsc.picwidth;
	bbox.bottom=dsc.picheight;

	if(bbox.right+dsc.Rand.left+dsc.Rand.right>paper.right
	|| bbox.bottom+dsc.Rand.top+dsc.Rand.bottom>paper.bottom){
		throw tgfError::error("Drawing does not fit on one page\n");	
	}

	of->write_head(paper,bbox,dsc.thickness);

	if(dsc.treetitle.length()>0){
		nfstring tit=nfstring(dsc.treetitle,"title");
		fPoint pt1=fPoint(dsc.Rand.left,dsc.Rand.top+dsc.getfontsize(tit));
		of->write_text(tit,pt1);
	}
	renderNode(this,of);

	renderRandklammern(of);
	
	if(dsc.uselength){
		fPoint p;
		p.x=dsc.rulepos.x;
		p.y=dsc.picheight-dsc.rulepos.y;
		
		float z=0.4*dsc.treewidth/dsc.lenscale;
		z=std::pow((float)10.0,(int)std::floor(log10(z)));
		//hoechstens 0.4mal Baumbreite als Lineal
		of->write_rule(p,z,z*dsc.lenscale);
	}

	of->write_foot();

}


void tree::renderNode(const tnode *cp, gostream *of) const
{

	std::string ts1,ob,un;
		
	//Text schreiben
	fPoint cc;
	if(!cp->root()){
		cc=cp->prev->pos;
		cc.y=cp->pos.y;
	} else {
		//sonst x-Koordinate des Vorgaengers  0
		cc=cp->pos;
		cc.x=0;
	}

	cc.x=0.5*(cc.x+cp->pos.x);



	fPoint c2=cc;
	//getfontsize liefert 0, wenn String leer
	float zh=std::max(dsc.getfontsize(cp->u[0]),dsc.getfontsize(cp->u[1]));
	c2.y-=0.33*zh;
	writeInnerNodeTextLine(c2,cp->u[0], cp->u[1], dsc.colsep, dsc,of);
	c2.y-=1.0*zh;
	writeInnerNodeTextLine(c2,cp->u[2], cp->u[3], dsc.colsep, dsc,of);
	
	c2=cc;
	zh=std::max(dsc.getfontsize(cp->d[0]),dsc.getfontsize(cp->d[1]));
	c2.y+=1.0*zh;
	writeInnerNodeTextLine(c2,cp->d[0], cp->d[1], dsc.colsep, dsc,of);
	zh=std::max(dsc.getfontsize(cp->d[2]),dsc.getfontsize(cp->d[3]));
	c2.y+=1.2*zh;
	writeInnerNodeTextLine(c2,cp->d[2], cp->d[3], dsc.colsep, dsc,of);
	

	if(dsc.shownumbers){
		fPoint p1=cp->pos;
		p1.x-=20;
		p1.y+=7;
		of->write_nodelabel(p1,cp->label());
	}

	if(cp->terminal()){//Randtext
		if(cp->r.length()==0){
			return;
		}
		cc=cp->pos;
		cc.x+=dsc.extrasep;
		cc.y+=0.33*dsc.getfontsize(cp->r);
		of->write_text(cp->r,cc);
		
		return;
	}	

	tnode *nx=cp->next;
	float xx=cp->pos.x;
	
	int nch=cp->nchildren();
	if(nch==1){ //nur ein Nachfolgeknoten
		of->write_hline(xx, cp->pos.y, nx->pos.x-xx);
	} else { //mindestens zwei Nachfolgeknoten
		fPoint p=nx->pos;
		float x1=p.x-xx,y1=p.y;
		while((nx=nx->down)) p=nx->pos;
		if(dsc.roundness==0)
			of->write_egabel(x1,y1,p.x-xx,p.y,xx);
		else
			of->write_rgabel(x1,y1,p.x-xx,p.y,xx,dsc.roundness*dsc.maxrund);

		nx=(cp->next)->down;//existiert n.V.
		while(nx->down){
			p=nx->pos;
			of->write_hline(xx, p.y, p.x-xx);
			nx=nx->down;
		}
	

	}
	
	//rekursiv weiter
	nx=cp->next;
	while(nx){
		renderNode(nx,of);
		nx=nx->down;
	}
	
}


void tree::renderRandklammern(gostream *of) const
{
  	for(std::list<klammer>::const_iterator it=kl.begin(); it!=kl.end(); ++it){
		const tnode *x[2];
		bool gut=true;
		for(int i=0; i<2; i++){

			if((x[i]=lb.bfind_label((*it).lb[i].c_str()))==0){
				std::cerr << "\nLabel '" << (*it).lb[i] << "' undefined.\n";
				gut=false;
			}
		}
		if(gut){
			if(x[0]->pos.y>x[1]->pos.y){
				const tnode *tmp=x[0]; x[0]=x[1]; x[1]=tmp;
			}
			x[0]=x[0]->oberster_sichtbarer();
			x[1]=x[1]->unterster_sichtbarer();
			float xpos=dsc.picwidth-dsc.Rand.right-dsc.klammerrand+(*it).sp*klammer::breite;
			of->write_klammer((*it).text,x[0]->pos.y,x[1]->pos.y,
				xpos,0.35*dsc.getstyle("r").fontsize,it->style);
		}
	}

}



void writeInnerNodeTextLine(const fPoint& p,const nfstring& s1, 
	const nfstring& s2,  const nfstring& sepw, const tree_desc& dsc, gostream *of)
{
	float x=p.x;
	float w=0,wd[3]={0,0,0};
	int nc=0;
	
	if(s1.size()>0){
		w=(wd[0]=dsc.helv_stringwidth(s1));
		++nc;
	}
	if(s2.size()>0){
		w+=(wd[2]=dsc.helv_stringwidth(s2));
		++nc;
	}
	if(nc==2){
		w+=(wd[1]=dsc.helv_stringwidth(sepw));
	}
	x-=w/2;

	fPoint p2=p;
	p2.x=x;
	if(s1.size()>0){
		of->write_text(s1,p2);
		p2.x+=wd[0];
	}
	if(s2.size()>0){
	 	of->write_text(sepw,p2);
		p2.x+=wd[1];
	 	of->write_text(s2,p2);
	}	
		
	


}


float min_node_width(const tnode *p,const std::string& colsep, const tree_desc& dsc)
{
	float sepw=dsc.helv_stringwidth(colsep);
	float m=0;
	for(int i=0; i<2; i++){
	//oben, i-te Zeile
		float m2=0;
		int nc=0;
		if(p->u[2*i].size()>0){
			m2=dsc.helv_stringwidth(p->u[2*i]);
			++nc;
		}
		if(p->u[2*i+1].size()>0){
			m2+=dsc.helv_stringwidth(p->u[2*i+1]);
			++nc;
		}
		if(nc==2) m2+=sepw;
		if(m2>m) m=m2;
	//unten, i-te Zeile	
		m2=0;
		nc=0;
		if(p->d[2*i].size()>0){
			m2=dsc.helv_stringwidth(p->d[2*i]);
			++nc;
		}
		if(p->d[2*i+1].size()>0){
			m2+=dsc.helv_stringwidth(p->d[2*i+1]);
			++nc;
		}
		if(nc==2) m2+=sepw;
		if(m2>m) m=m2;
	}
	
	const float extra=2.5*millimeter;
	return m+extra; 

}


