/*
	tgf-tree.cc, part of
	
	treegraph
	
	Tree formatting program
	
	Generates vector graphics (SVG,EPS) from .tgf-tree description files.

	Copyright (c) 2003-06 by Joern Mueller


	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program (GPL.html); if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
	(also: http://www.gnu.org)
*/



#include "tgf-main.h"
#include "tgf-tree.h"
#include "tgf-render.h"
#include "tgf-error.h"
#include <cstdlib>
#include <cstring>

using namespace std;

float klammer::breite=40.0;



int bound(const char *s, tnode **ls, int n);

void astbreiten(vector<float>& a, int n, float w);
float min_astbreite(const vector<float>& a, int n, float w);

tnode::tnode(tnode *prv, float l) : len(l), prev(prv)
{
	next=down=0;
	lbl=0;
	prune=false;
	userset[0]=userset[1]=userset[2]=false;
	minlen=minabove=minbelow=autolength;
}



void tnode::setlabel(const string &a)
{
	if(has_label()) delete [] lbl;
	lbl=new char [a.length()+1];
	strncpy(lbl,a.c_str(),a.length()+1);
}

int tnode::nchildren() const
{

	int nc=0;
	
	tnode *p=next;
	while(p){
		++nc;
		p=p->down;
	}
	return nc;
}

int tnode::nterminals() const
{
	if(terminal()) return 1;
	int nt=0;
	tnode *p=next;
	while(p){
		nt+=p->nterminals();
		p=p->down;
	}
	return nt;
	

}


void tnode::calcdepth_sichtbar()
{
	depth=0;
	if(!prune){
		tnode *p=next;
		while(p){
			p->calcdepth_sichtbar();
			if(p->depth+1>depth)
				depth=p->depth+1;
			p=p->down;
		}
	}
}


void tnode::selection_sort_children(int sgn)
{
	int nn=nchildren();
	if(nn<2) return;
	
	tnode **nl=new tnode* [nn];
	nl[0]=next;
	int i=1;
	for(; i<nn; i++)
		nl[i]=nl[i-1]->down;
//selection sort
	for(i=0; i<nn-1; i++){
		int mn=i;
		for(int j=i+1;j<nn;j++)
			if(sgn*(nl[j]->depth-nl[mn]->depth)>0) mn=j;
		tnode *tmp=nl[mn]; nl[mn]=nl[i]; nl[i]=tmp;
	}		
//Verbindungen richtig machen
	next=nl[0];
	for(i=1; i<nn; i++){
		nl[i-1]->down=nl[i];
	}
	nl[nn-1]->down=0;
		
	delete [] nl;
}

bool tnode::permute(list<int>& perm)
{
// Testen, ob wirklich Permutation von {1,..,nchildren} vorliegt
	int nn=nchildren();
	if(nn<2 || static_cast<int>(perm.size())!=nn) return false;
	list<int>::const_iterator iit=perm.begin();
	int mi= *iit;
	int ma=mi;
	while(++iit!=perm.end()){
		int nu= *iit;
		if(nu>ma) ma= nu;
		else if(nu<mi) mi= nu;
	}
	if(mi!=1 || ma!=nn)
		return false; //keine Permutation
	
	tnode **nl=new tnode* [nn];
	tnode *p=next;

	for(int j=0;j<nn;j++){
		nl[j]=p;
		p=p->down;
	}

	iit=perm.begin();
	next=nl[(*iit)-1];
	p=next;
	++iit;
	while(iit!=perm.end()){
		p->down=nl[(*iit)-1];
		p=p->down;
		++iit;
	}
	p->down=0;

		
	delete [] nl;
	return true;
}


void tnode::swap_children()
{
	int nn=nchildren();
	if(nn!=2){
		cout << "Warning: tnode::swap_children: Node must have exactly 2 children.\n";
		return;
	}
	
	tnode *tmp1=next;
	tnode *tmp2=next->down;
	
	tmp1->down=0;
	tmp2->down=tmp1;
	next=tmp2;
}

void tnode::collapse()
{
	if(prev==0) return;
	tnode *adown=down;
	down=next;
	tnode *t=next;
	tnode *zt=t;
	while(t){
		t->prev=prev;
		zt=t;
		t=t->down;
		
	}
	zt->down=adown;

	next=0;
	prev->removechild(this);
}


void tnode::ladderize_sichtbare(int d)
{
	if(prune) return;

	selection_sort_children(d);
	tnode *p=next;
	while(p){
		p->ladderize_sichtbare(d);
		p=p->down;
	}	
}

void tnode::symladderize(int d)
{
	
	selection_sort_children(d);
	int i=1;
	int n=nchildren();
// Permutation n-1 n-3 .. > < .. n-2 n  anwenden
	if(n>=3){
		list<int> il;
		list<int>::iterator it=il.end();
		for(i=0; i<n; i++){
			il.insert(it,n-i);
			if(i%2==0) --it;
		
		}
		permute(il);
	} else if(n==1){
		next->symladderize(d);
		return;
	}
	i=0;
	tnode *p=next;
	if(prune) return;
	while(p){
		p->symladderize(i<n/2?1:-1);
		++i;
		p=p->down;
	}	
}


const tnode *tnode::unterster_sichtbarer() const
{
	const tnode *u=this;
	while(u->next && !u->prune){
		u=u->next;
		while(u->down) u=u->down;
	}
	return u;
}

const tnode *tnode::oberster_sichtbarer() const
{
	const tnode *u=this;
	while(u->next && !u->prune){
		u=u->next;
	}
	return u;
}

void tnode::removechild(tnode *ch)
//setzt voraus, dass ch als child existiert, insbesondere nchildren()>0
{

	if(next==ch){
		next=ch->down;
	}else {
		tnode *p=next;
		while(p->down){
			if(p->down==ch){
				p->down=ch->down;
				break;
			}
			p=p->down;
		}
	}	
	ch->prev=0;
	ch->down=0;
}


void tnode::insertchild(tnode *ch)
// setzt ch->down=ch->prev=0 voraus
{	
	ch->down=next;
	next=ch;
	ch->prev=this;
		
}



label_list::label_list()
{
	nd=0;
	nnodes=nlabel=0;
}

label_list::~label_list()
{
	if(nd) delete [] nd;
}


void label_list::init(int a)
{
	if(nd) delete [] nd;
	nd = new tnode* [nnodes=a];
	for(int i=0; i<nnodes; i++) nd[i]=0;
	nlabel=0;
	
	autovar=1;//fuer auto_label
}

int label_namecompare(const char *s1, const tnode *p2)
{
	return strcmp(s1,p2->label());
}


void label_list::einsortiere_label(tnode *p)
{
	if(nlabel<1){
		nd[0]=p;
		nlabel=1;
		return;
	}
		
	if(nlabel>=nnodes){
		throw tgfError::error("Label buffer overflow");
	}
	int nb=bound(p->label(),nd,nlabel);
	for(int i=nlabel-1; i>nb; i--)
		nd[i+1]=nd[i];
	nd[nb+1]=p;
	++nlabel;
	if(nb>=0 && label_namecompare(p->label(),nd[nb])==0){
		throw tgfError::error("Duplicate label '" + std::string(p->label()) + "'");
	}
}

void label_list::kill_label(const char *s)
{
	if(nlabel<1){
		return;
	}
		

	int nb=bound(s,nd,nlabel);
	if(nb<0) return;
	if(label_namecompare(s,nd[nb])!=0)
		return;
		
	for(int i=nb; i<nlabel-1; i++)
		nd[i]=nd[i+1];
	--nlabel;
}



void label_list::random_label(char s[])
{
	char tmp[12];
	do {
		sprintf(tmp,"n%d",autovar++);
	}while(bfind_label(tmp));
	strncpy(s,tmp,12);	
}


int bound(const char *s, tnode **ls, int n)
//groesstes i, 0<=i<n, so dass ls[i]<=s, oder -1, falls s<ls[0]
{
	if(n<1) return 0;
	
	if(label_namecompare(s,ls[0])<0) return -1;
	int l=0,h=n;
	while(h-l>1){
		int m=(h+l)/2;
		if(label_namecompare(s,ls[m])<0) h=m;
		else l=m;
	}
	return l;

}


const tnode *label_list::bfind_label(const char *s) const
{
	int nb=bound(s,nd,nlabel);
	if(nb<0) return 0;
	if(label_namecompare(s,nd[nb])!=0)
		return 0;
	return nd[nb];
}


float tnode::maxrund() const
//erst nach calcpos aufrufen!
{
	if(terminal()) return -1.0;
	if(nchildren()==1) return next->maxrund();
	
	tnode *p=next;
	tnode *p2=p->down;
	float mr=min((float)abs(p2->pos.y-p->pos.y),p->pos.x-pos.x);
	mr=min((float)abs(pos.y-p->pos.y),mr);
	while(p2->down){
		p=p2;
		p2=p2->down;
	}
	mr=min((float)abs(p2->pos.y-p->pos.y),mr);
	mr=min((float)abs(pos.y-p2->pos.y),mr);
	
	p=next;
	while(p){
		float r2=p->maxrund();
		if(r2>0 && r2<mr) mr=r2;
		p=p->down;
	}
	return mr;

}


tree::tree() : tnode(0)
{
	drty=true;
}

tree::~tree()
{
	if(has_label()) lb.kill_label(label());
	remove_subtree(this);
}


void tree::read_definitions(tgfParser& P)
//alles zwischen (Dateianfang) und \enddef
{
	int mode=0;
		
	dsc.Rand=fRect();
	string wort;	
	definitionitems.clear();
	int ch;
	do{
		ch=P.readnextword(wort,mode==1);
		bool ok=false;
		if(ch==tgfParser::kEOF){
			P.fehler("Unexpected end of file");			
		} else if(mode==0 && ch==tgfParser::kBeginDef){
			mode=1;
			ok=true;
		} else if(mode==1) {
			if(ch==tgfParser::kCommand){
				bool normal=true;
				if(wort=="width"){ 
					dsc.picwidth=P.readbracefloat()*millimeter;
				}  else if(wort=="height"){
					dsc.picheight=P.readbracefloat()*millimeter;
				} else if(wort=="autolength"){
					dsc.forceauto=true;
				} else if(wort=="style"){
					string ident,typ;
					P.readbracestring(ident);
					// for compatibility with old versions: --> 
					if(ident=="u") ident="u1";
					else if(ident=="d") ident="d1";
					// <--
					P.readbracestring(typ);
					float sz=P.readbracefloat();
					dsc.setstyle(ident,fontstyle(typ,sz));
					if(ident=="r"){
						dsc.extrasep=dsc.helv_stringwidth(nfstring("x","r"));
					}
					normal=false;
					ok=true;
					definitionitems.push_back("#"+ident);
				} else if(wort=="thickness"){
					dsc.thickness=P.readbracefloat()*millimeter;
			//	} else if(wort=="scale"){
			//		dsc.scale=P.readbracefloat();
				} else if(wort=="variable"){
					dsc.variablelength=true;
				} else if(wort=="proof"){
					dsc.shownumbers=true;
				} else if(wort=="roundness"){
					dsc.roundness=P.readbracefloat();
				} else if(wort=="rulepos"){
					dsc.rulepos.x=P.readbracefloat()*millimeter;
					dsc.rulepos.y=P.readbracefloat()*millimeter;
				} else if(strncmp(wort.c_str(),"brac",4)==0) { 
					//Randklammer: (\brace|\bracket|\brace*|\bracket*){spalte}{lb1}{lb2}{name}
					klammer nk;
					if(strncmp(wort.c_str(),"brace",5)==0)
						nk.style=klammer::geschweift;
					else nk.style=klammer::eckig;
					if(wort[wort.length()-1]=='*') nk.style+=klammer::gedreht;
					nk.sp=P.readbracefloat();
					P.readbracestring(nk.lb[0]);
					P.readbracestring(nk.lb[1]);
					string kltxt;
					P.readbracestring(kltxt);
					nk.text=nfstring(kltxt,"br");
					kl.push_back(nk);
					
					normal=false;
					ok=true;
					definitionitems.push_back("brac");
	
				} else if(wort=="margin") {
					dsc.Rand.left=P.readbracefloat()*millimeter;
					dsc.Rand.top=P.readbracefloat()*millimeter;
					dsc.Rand.right=P.readbracefloat()*millimeter;
					dsc.Rand.bottom=P.readbracefloat()*millimeter;
				} else if(wort=="paper"){
					P.readbracestring(dsc.paper);
				} else if(wort=="landscape"){
					dsc.landscape=true;
				} else if(wort=="separator"){
					P.readbracestring(dsc.colsep);
				} else if(wort=="version"){
					dsc.tgfversion=P.readbracefloat();
					if(dsc.tgfversion>kTGFfileformat){
						cerr << "WARNING: File has been generated by a newer version\n";
						cerr << "of treegraph.\nPlease update to the latest version.\n";
					}
				} else if(wort=="title"){
					P.readbracestring(dsc.treetitle);
				} else if(wort=="name"){
					P.readbracestring(dsc.treename);
				} else if(wort=="font"){
					
					cerr << "Warning: Different fonts not yet supported.\n";
					normal=false;
				} else {
					cerr << "Warning: Unknown command \\" << wort << " in line "<<P.curline()<<".\n";
					int arg=0;
					while(P.skipbracestring()){++arg;}
					if(arg>0){
						cerr << "--> Ignored ";
						if(arg==1) cerr << "argument";
						else cerr << arg << " arguments";
						cerr << " of \\" << wort << ".\n";
						
					}
					normal=false;
					ok=true;
				}
				if(normal){
					definitionitems.push_back(wort);
					ok=true;
				}
				
			} else if(ch==tgfParser::kEndDef){
					break;
			} else if(ch==tgfParser::kNewline || ch==tgfParser::kComment){
				definitionitems.push_back(wort);
				ok=true;
			}

		} //mode==1
		if(!ok){
			P.fehler("Illegal character '"+tgfError::CharToString(ch)+"' (ASCII "+
					tgfError::IntegerToString(ch)+")");
		}
	} while(!P.eof() && P.good());
	

}


void tree::read(ifstream& input)
// alles zwischen '(' und ')' einlesen
{

	dsc.uselength=dsc.forceauto=false;

	tgfParser P(input);
	read_definitions(P);
	
	tnode *cp=this;
	
	int knoten=1;
	
	tgfParser::ttoken lastt=tgfParser::kInit;
	string plaintxt="";//Position noch nicht zugeordnet
	while(lastt!=tgfParser::kEOF){
		string s;
		tgfParser::ttoken was=P.readnextword(s);
		
		if(cp==0){
			P.fehler("tree::read::1");
		}

		switch(was){
		case tgfParser::kNextTree:	
			cp->down=new tnode(cp->prev);
			++knoten;
			if(plaintxt.length()>0){
				if(!(cp->next))
					cp->r=nfstring(plaintxt,"r");
				else // ".." war hinter child node
					cp->u[0]=nfstring(plaintxt,"u1");
				plaintxt="";
			}
			cp=cp->down;
			break;
		case tgfParser::kOpenTree:
			cp->next=new tnode(cp);
			++knoten;
			if(plaintxt.length()>0){
				cp->u[0]=nfstring(plaintxt,"u1");
				plaintxt="";
			}
			cp=cp->next;
			break;
		case tgfParser::kCloseTree:
			if(plaintxt.length()>0){
				if(!(cp->next))
					cp->r=nfstring(plaintxt,"r");
				else // ".." war hinter child node
					cp->u[0]=nfstring(plaintxt,"u1");
				plaintxt="";
			}
			cp=cp->prev;
			break;
		case tgfParser::kText:
			plaintxt=s;
			break;
		case tgfParser::kCommand:
			if(s=="r"){
				P.readbracestring(s);
				cp->r=nfstring(s,"r");
			} else if(s.size()==2 && (s[0]=='u' || s[0]=='d') && (s[1]>='1' && s[1]<='4')){
				string arg;
				P.readbracestring(arg);
				int p=s[1]-'1';
				if(s[0]=='u')
					cp->u[p]=nfstring(arg,s);
				else
					cp->d[p]=nfstring(arg,s);
		
			} else if(s=="len") {	
				cp->len=P.readbracefloat();
				dsc.uselength=true;
			} else if(s=="label") {	
				P.readbracestring(s);
				cp->setlabel(s);
			} else if(s=="prune") {	
				cp->prune=true;
			} else if(s=="min_length") {
				cp->minlen=P.readbracefloat()*millimeter;
				cp->userset[0]=true;
			} else if(s=="space_above") {
				cp->minabove=P.readbracefloat()*millimeter;
				cp->userset[1]=true;
			} else if(s=="space_below") {
				cp->minbelow=P.readbracefloat()*millimeter;
				cp->userset[2]=true;
			} else {
				cerr << "Warning: Unknown command \\" << s << " in line "<<P.curline()<<".\n";
				int arg=0;
				while(P.skipbracestring()){++arg;}
				if(arg>0){
					cerr << "--> Ignored ";
					if(arg==1) cerr << "argument";
					else cerr << arg << " arguments";
					cerr << " of \\" << s << ".\n";
				}
			}
			break;
		case tgfParser::kEOF:
			break;
		default:
			P.fehler("Unexpected token "+tgfError::IntegerToString(was)+" in tree::read");
			break;
		}
		
		
		lastt=was;
	}


	if(dsc.forceauto)
		dsc.uselength=false;
		
	if(dsc.uselength){
		if(dsc.roundness > 0){
			cout << "Warning: Rounded corners are not allowed in this mode.\n";
			cout << "--> Using \\roundness{0}.\n";
			dsc.roundness=0;
		}
	
	}
	
	cout << "Total number of nodes: " << knoten << endl;

	drty=false;
	fill_labels(knoten);
	

	calcpos();
		
}

void tree::write(ostream& aus, const tnode *tn)
{
	
	//Definition-part
	aus << "\\begindef ";
	list<string>::const_iterator cil=definitionitems.begin();
	list<klammer>::const_iterator ckl=kl.begin();
	while(cil!=definitionitems.end()){
		string s= *cil;
		++cil;
		if(s[0]=='\n' || s[0]=='%'){//newlines, comments
			aus<< s;
			continue;
		}
		aus << '\t';
		if(s[0]=='#'){//style
			string typ(s,1,string::npos);
			if(typ=="u") typ="u1";
			else if(typ=="d") typ="d1";
			tree_desc::shownamedstyle(aus,namedfontstyle(typ,dsc.getstyle(typ)));
			aus << ' ';
		} else if(s=="brac"){
			//in der Hoffnung, dass sich Reihenfolge der Klammern
			//seit Einlesen nicht geaendert hat..
			if(ckl!=kl.end()){
				const klammer& K= *ckl;
				if(K.style & klammer::eckig){
					aus << "\\bracket";
				} else {
					aus << "\\brace";
				}
				if(K.style & klammer::gedreht){
					aus << '*';
				}
				aus << '{' << K.sp << "}{" << K.lb[0] << "}{" <<K.lb[1] << "}{"<<K.text<<"} ";
				++ckl;
			}	
		} else if(s=="width"){ 
			aus << "\\width{"<<dsc.picwidth/millimeter<<"} ";
		}  else if(s=="height"){
			aus << "\\height{"<<dsc.picheight/millimeter<<"} ";
		} else if(s=="autolength"){
			aus <<"\\autolength";
		} else if(s=="thickness"){
			aus << "\\thickness{"<<dsc.thickness/millimeter<<"} ";
		} else if(s=="variable"){
			aus << "\\variable";
		} else if(s=="proof"){
			aus << "\\proof";
		} else if(s=="roundness"){
			aus << "\\roundness{"<<dsc.roundness<<"} ";
		} else if(s=="margin") {
			aus << "\\margin{";
			aus << dsc.Rand.left/millimeter<<"}{";
			aus << dsc.Rand.top/millimeter<<"}{";
			aus << dsc.Rand.right/millimeter<<"}{";
			aus << dsc.Rand.bottom/millimeter<<"} ";
		} else if(s=="paper"){
			aus << "\\paper{"<<dsc.paper<<"} ";
		} else if(s=="landscape"){
			aus << "\\landscape ";
		} else if(s=="separator"){
			aus << "\\separator{"<<dsc.colsep<<"} ";
		} else if(s=="rulepos"){
			aus << "\\rulepos{"<<dsc.rulepos.x/millimeter <<"}{"<< dsc.rulepos.y/millimeter<<"}\n";
		} else if(s=="name"){
			aus << "\\name{" <<dsc.treename<<"} ";
		} else if(s=="version"){
			aus << "\\version{"<<dsc.tgfversion<<"} ";
		} else {
			aus<< "#ERROR!!!\n";
		}
		
		
	}
	aus << "\\enddef\n";
	
	
	//Baum	
	writenodes(aus,tn);
	
	
	if(tn==0) drty=false; //ganzer Baum wird gespeichert
	
	flush(aus);
}


void tree::writenodes(ostream& aus, const tnode *tn) const
{
	const tnode *nd=(tn==0)?this:tn;
	const tnode *here=nd;

	const int maxindent=25;
	
	char *sp=new char[maxindent+3];
	int i;
	for(i=0; i<maxindent+2; i++)
		sp[i]=' ';
	sp[maxindent+2]=0;
	sp[0]='\n';
	sp[1]=0;
	
	int level=0;
	do{

		bool something=false;
		if(nd->len!=autolength){
			aus << sp <<"\\len{" << nd->len << '}';
			something=true;
		}
		
			
		if(nd->r.length()>0){
			aus << (something?" ":sp);
			aus << "\\r{" << nd->r << '}';
			something=true;
		}
		for(i=0; i<4; i++){
			if(nd->u[i].length()>0){
				aus << (something?" ":sp);
				aus << "\\u"<<i+1<<'{' << nd->u[i] << '}';
				something=true;
			}
			
		}
		for(i=0; i<4; i++){
			if(nd->d[i].length()>0){
				aus << (something?" ":sp);
				aus << "\\d"<<i+1<<'{' << nd->d[i] << '}';
				something=true;
			}
			
		}
		
		if(nd->minlen>=0 && nd->userset[0]){//Auch explizites 0 soll erlaubt sein
			aus << (something?" ":sp);
			aus << "\\min_length{" << nd->minlen/millimeter << '}';
			something=true;
		}
		if(nd->minabove>=0 && nd->userset[1]){
			aus << (something?" ":sp);
			aus << "\\space_above{" << nd->minabove/millimeter << '}';
			something=true;
		}
		if(nd->minbelow>=0 && nd->userset[2]){
			aus << (something?" ":sp);
			aus << "\\space_below{" << nd->minbelow/millimeter << '}';
			something=true;
		}
		
		
		if(nd->prune){
			aus << "\\prune";
		}
		
		if(nd->has_label()){
			aus << (something?" ":sp);
			aus << "\\label{" << nd->label() << '}';
			something=true;
		}
		int l;
		if(nd->next){
			aus << sp << '(';
			++level;
			for(l=1; l<min(level,maxindent)+1; l++) sp[l]=' ';
			sp[0]='\n'; sp[l]=0;
			nd=nd->next;
		} else {
			int j=0;
			while(!nd->down && nd!=here){
				nd=nd->prev;
				++j;
			}
			level-=j;
			for(l=1; l<min(level,maxindent)+1; l++) sp[l]=' ';
			sp[0]='\n'; sp[l]=0;
			if(j>0){
				aus << sp;
				while(j-- >0) aus << ')';
			}
			if(nd!=here){
				nd=nd->down;
				aus << ',';
			}
		}
	} while(nd!=here);
	aus << endl;
	delete [] sp;
}


void tree::writeNEXUS(ostream& aus)
{
	aus << "#NEXUS\n\nbegin trees;\ntree TheTree=[&u]";
	
	writenodesNEXUS(aus);
	
	aus << ";\nend;\n";
	
	flush(aus);
}

void tree::writenodesNEXUS(ostream& aus, const tnode *tn) const
{
	const tnode *nd=(tn==0)?this:tn;
	

	if(nd->next){
		aus << '(';
		nd=nd->next;
		while(nd){
			writenodesNEXUS(aus,nd);
			nd=nd->down;
			if(nd){
				aus<<',';
			}
		}
		aus << ')';
	} else {
		if(nd->r.length()>0){
			string tmp="";
			tmp.reserve(nd->r.length()+1);
			for(int i=0; i<static_cast<int>(nd->r.length()); i++){
				char ch=nd->r[i];
				if(!isalnum(ch)) tmp.push_back('_');
        			else tmp.push_back(nd->r[i]);
			}
			aus  << tmp;
		}
	}
	if(tn!=0){
		if(tn->len!=autolength)
			aus << ':' << tn->len;
	}
	
	
}




void tree::fill_labels(int nnodes1)
{
	lb.init(nnodes1);
	setlabel("root"); //ggf. user-label ueberschreiben
	
	user_labels(this,lb);
	auto_labels(this,lb);
	
}




void tree::user_labels(tnode *q, label_list& lbl)
{
	if(q->has_label()){
		lbl.einsortiere_label(q);
	}
	tnode *p=q->next;
	while(p){
		user_labels(p,lbl);
		p=p->down;
	}

}

void tree::auto_labels(tnode *q, label_list& lbl)
{
	if(!q->has_label()){
		char tmp[12];
		lbl.random_label(tmp);
		q->setlabel(tmp);
		lbl.einsortiere_label(q);
		if(!drty){
			cout << "Note: Consecutively numbered labels have been added\n";  
			cout <<  "     where labels were missing\n"; 
		}
		drty=true;
	}
	tnode *p=q->next;
	while(p){
		auto_labels(p,lbl);
		p=p->down;
	}
}


void tree::remove_subtree(tnode *p)
// alle children von p loeschen
{
	tnode *q=p;
	p=p->next;
	while(p){
		if(p->has_label()){
			lb.kill_label(p->label());
			delete [] p->lbl;
		}
		remove_subtree(p);
		p=p->down;
	}
	q->next=0;

}



void tree::calc_mindimensions()
//minimale Astlaengen und minimale Baumbreite bestimmen
{
	dsc.min_treewidth=0;
	dsc.min_treeheight=0;
	float nw=0;
	tnode *p=this;
	if(minlen<=0) minlen=0;//root hat Laenge 0
	for(;;){
		while(!p->terminal()){
			if(p->minlen==autolength)
				p->minlen=min_node_width(p,dsc.colsep,dsc);
			nw+=p->minlen;
			p=p->next;
		}
		//Randknoten
		if(p->minlen==autolength)
			p->minlen=min_node_width(p,dsc.colsep,dsc);
		nw+=p->minlen;
		if(nw>dsc.min_treewidth) dsc.min_treewidth=nw;
		
		float rs=(float)0.6*dsc.getstyle("r").fontsize;
		dsc.min_treeheight+=max(p->minabove,rs);
		dsc.min_treeheight+=max(p->minbelow,rs);
		
		while(p->down==0){
			nw-=p->minlen;
			p=p->prev;
			if(p==0){
				return;
			}
		}
		nw-=p->minlen;
		p=p->down;
	}
}

void tree::calc_treewidth()
//Maximal moegliche Breite bestimmen
{
//Randklammern ausmessen

	dsc.klammerrand=0;
	for(list<klammer>::const_iterator it=kl.begin(); it!=kl.end(); ++it){
		float wd=(*it).sp*klammer::breite;//kann auch <0 sein

		if(it->style & klammer::gedreht)
			wd+=dsc.helv_stringwidth(nfstring(it->text,"br"))+12.0;//12:klammer selbst
		else
			wd+=klammer::breite;
		if(wd>dsc.klammerrand) dsc.klammerrand=wd;
		
	}


//Randtexte ausmessen
	if(dsc.uselength){
		dsc.treewidth=dsc.picwidth-dsc.Rand.left-dsc.Rand.right//
			-2*dsc.extrasep-dsc.klammerrand;
		return;
	}
	
	
	
	float randw=0;
	tnode *p=this;
	for(;;){
		while(p->next && !p->prune){
			p=p->next;
		}
		//Randknoten erreicht
		float textb=dsc.helv_stringwidth(p->r);
		if(textb>randw) randw=textb;
		
		while(p->down==0){
			p=p->prev; 
			if(p==0){
				dsc.treewidth=dsc.picwidth-dsc.Rand.left//
					-dsc.Rand.right-randw-2*dsc.extrasep
					-dsc.klammerrand;
				return;
			}
		}
		p=p->down;
	}
}


inline float zentrum(float a, float b, float c)
//zweitgroesster Wert von a,b,c
//Voraussetzung: a<c
{
	if(b<=a) return a;
	else if(b>=c) return c;
	
	return b;
}

void tree::calcvpos()
//vertikal-Koordinaten
// beruecksichtigt space_above/below nur bei terminal-Knoten
{
	int nt=nterminals();
	vector<float> ht(nt);
	vector<fPoint> pt(nt); //x->above,y->below
//minimale hoehen der Randknoten sammeln
	int i=0;
	tnode *p=this;
	for(;;){
		while(!p->terminal()) p=p->next;
		float rs=(float)0.6*dsc.getstyle("r").fontsize;
		pt[i].x=max(p->minabove,rs);
		pt[i].y=max(p->minbelow,rs);
		ht[i]=pt[i].x+pt[i].y;
		i++;
		
		while(p->down==0){
			p=p->prev;
			if(p==0) break;
		}
		if(p) p=p->down;
		else break;
	}
//optimale hoehen der Randknoten berechnen.	
	astbreiten(ht,nt,dsc.picheight-dsc.Rand.top-dsc.Rand.bottom);

//vertikale Positionen
	p=this;
	i=0;
	float v1=0;
	float v0=0;
	for(;;){
		while(p->next && !p->prune) p=p->next;
		
		//Position moeglichst zentral in [v0,v0+ht[i]]
		p->pos.y=v0+zentrum(pt[i].x,ht[i]/2,ht[i]-pt[i].y);
		v0+=ht[i];
		i++;
		while(p->down==0){
			v1=p->pos.y;
			p=p->prev;
			if(p==0) return;
			p->pos.y=0.5*(p->next->pos.y+v1);
		}
		p=p->down;
	}


}

struct _fi
{
	float f;
	int i;
};

int _ficomp(const void *a, const void *b)
{
	float d=((_fi*)a)->f-((_fi*)b)->f;
	return (int)d;
}

int _floatcomp(const void *a, const void *b)
{
	float d=*((float*)a)- *((float*)b);
	return (int)d;
}


void astbreiten(vector<float>& a, int n, float w)
{
	if(static_cast<unsigned int>(n)>a.size()){//may not happen
		throw tgfError::error("astbreiten[1]");
	}


	_fi *b=new _fi[n];
	int i;
	for(i=0;i<n; i++){
		b[i].i=i;
		if(a[i]<0) b[i].f=0;
		else b[i].f=a[i];
	}
	qsort(b,n,sizeof(_fi),_ficomp);
	
	float r=w/((float)n);
	for(i=n-1; i>=0; i--){
		if(b[i].f>r){
			w-=b[i].f;
			if(i==0){
				delete [] b;
				throw tgfError::error("astbreiten(): "+tgfError::FloatToString(-w/millimeter)+"mm too wide");
			}
			r=w/((float)i);
		} else b[i].f=r;
		
	}
	
	for(i=0;i<n; i++){
		a[b[i].i]=b[i].f;
	}
	delete [] b;
	

}

float min_astbreite(const vector<float>& a, int n, float w)
{
	if(static_cast<unsigned int>(n)>a.size()){//may  not happen
		throw tgfError::error("min_astbreite[1]");
	}


	float *b=new float [n];
	int i;
	for(i=0;i<n; i++){
		if(a[i]<0) b[i]=0;
		else b[i]=a[i];
	}
	qsort(b,n,sizeof(float),_floatcomp);
	
	float r=w/((float)n);
	for(i=n-1; i>=0; i--){
		if(b[i]>r){
			w-=b[i];
			if(i==0){
				delete [] b;
				throw tgfError::error("min_astbreite() "+tgfError::FloatToString(-w/millimeter)+"mm too wide");
			}
			r=w/((float)i);
		} else b[i]=r;
		
	}
	
	float f=b[0];
	delete [] b;
	
	return f;
}



int tree::calchpos()
//automatisch, gleichmaessig verteilte Astlaenge (\variable)
{

	
	vector<float> ml(depth+1,0.0); //depth von root 1 zu klein
//
// Hier kann nicht einfach
//	sethpos(dsc.Rand.left,dsc.treewidth);
// aufgerufen werden, weil root immer Laenge 0 hat.
// stattdessen fuer alle root-Nachfolger.
		
	pos.x=dsc.Rand.left; //root
	tnode *p=next;
	if(!p || prune){
		return 1; //nur ein root-Knoten
	}
	while(p){
		p->sethpos(pos.x,dsc.treewidth);
		p=p->down;
	}
	
	return 1;	
}


float tnode::optlen(vector<float>& x,int tief,float w)
{
	if(static_cast<unsigned int>(tief)>=x.size()){//may not happen
		throw tgfError::error("optlen[1]");
	}
	x[tief]=minlen;
	
	tnode *q=next;
	if(q==0 || prune){
		vector<float> ls=x;
		astbreiten(ls,tief+1,w);
		return ls[0];
	}

	float ol=q->optlen(x,tief+1,w);
	while(q->down){
		q=q->down;
		float ol1=q->optlen(x,tief+1,w);
		if(ol1<ol) ol=ol1;
	}
	return ol;
}


void tnode::sethpos(float xprev,float w)
{
	if(w<=0){
		throw tgfError::error("tnode::sethpos: no space for node");
	}
	vector<float> x(depth+1,0.0); //an der root-Tiefe soll's nicht scheitern
	float opt=optlen(x,0,w);
	pos.x=xprev+opt;
	w-=opt;

	tnode *q=next;
	if(q==0 || prune){
		return;
	}
	q->sethpos(pos.x,w);
	while(q->down){
		q=q->down;
		q->sethpos(pos.x,w);
	}
}



int tree::calchpos_fixed()
//automatisch, aber mit fester Standardlaenge
{

	
	vector<float> ml(depth+1,0.0); //depth von root 1 zu klein

//kleinste tatsaechliche Laenge berechnen
	pos.x=dsc.Rand.left; //root
	tnode *p=next;
	if(!p || prune){
		return 1; //nur ein root-Knoten
	}
	float fl=dsc.treewidth;	
	while(p){
		float f10=p->normlen_fixed(ml,0,dsc.treewidth);
		if(f10<fl) fl=f10;
		p=p->down;
	}
	
	p=next;
	while(p){
		p->sethpos_fixed(dsc.Rand.left,dsc.Rand.left+dsc.treewidth,fl);
		p=p->down;
	}
	
	
	return 1;	
}


float tnode::normlen_fixed(vector<float>& x,int tief,float w)
{
	if(static_cast<unsigned int>(tief)>=x.size()){//may not happen
		throw tgfError::error("normlen_fixed[1]");
	}
	x[tief]=minlen;
	
	tnode *q=next;
	if(terminal()){
		vector<float> ls=x;
		return min_astbreite(ls,tief+1,w);
	}

	float ol=q->normlen_fixed(x,tief+1,w);
	while(q->down){
		q=q->down;
		float ol1=q->normlen_fixed(x,tief+1,w);
		if(ol1<ol) ol=ol1;
	}
	return ol;
}


float tnode::sethpos_fixed(float x0,float w,float stdw)
{
	
	float l=stdw;
	if(minlen>l) l=minlen;
	tnode *q=next;
	if(q==0 || prune){
		pos.x=w;
		return pos.x-l;
	}
	
	float minh= q->sethpos_fixed(x0,w,stdw);
	while(q->down){
		q=q->down;
		float f2=q->sethpos_fixed(x0,w,stdw);
		if(f2<minh) minh=f2;
	}
	pos.x=minh;
	if(pos.x<x0){
		throw tgfError::error("tnode::sethpos_fixed: no space for node");
	}
	return pos.x-l;
}



int tree::calchpos_len()
//Astlaengen aus \len
{
	if(terminal()){
		throw tgfError::error("tree::calchpos_len: Only one node (root)");
	}
	
//Noetige Skalierung der \len-Laengen ermitteln	
	
	float smin=0, smax=1e10;
	
	
	len=0; //root
	minlen=autolength; //wenn len==0, keine minlen>0 erlaubt
		
	getoptlenscale(0, smin, smax, dsc);
	if(smin>smax){
		float f=(smin/smax-1.0)*dsc.treewidth;
		cout << "Warning: Tree is " << f/millimeter << "mm too wide\n";
		cout << "-->  Ignoring width of node labels and \\min_length.\n";
	}
	
	sethpos_len(dsc.Rand.left,smax);
	dsc.lenscale=smax;
	
	return 1;	
}

void tnode::getoptlenscale(float zweigl, float& smin, float& smax, const tree_desc& dsc)
//smax durch Baumbreite, smin durch minlen
{
	if(len==autolength){
	 cerr << "Missing \\len in node '"+r+"' (Label "+label()+"). Assuming len 0.\n";
	 len=0;
	//	throw tgfError::error("Missing \\len in node '"+r+"' (Label "+label()+")");
	}
	if(minlen!=autolength && minlen>0){
		
		if(len==0){
			if(userset[0]){
				throw tgfError::error("Node with \\len{0} must not have positive \\min_length. Here: \\min_length{"+tgfError::FloatToString(minlen)+"}");
			} else minlen=0;
		} else {
			float f=minlen/len;
			if(f>smin) smin=f;
		}
	}
	zweigl+=len;
	if(terminal()) {
	//Achtung. Bei prune muss dann Text auch in r stehen
		float f=(dsc.treewidth-dsc.helv_stringwidth(this->r))/zweigl;
		if(f<smax) smax=f;
	} else {
		tnode *p=next;
		while(p){
			p->getoptlenscale(zweigl,smin,smax,dsc);
			p=p->down;
		}
	}
	

}

void tnode::sethpos_len(float x0,float sc)
{
	pos.x=x0+sc*len;
	if(terminal()) return;
	tnode *p=next;
	while(p){
		p->sethpos_len(pos.x,sc);
		p=p->down;
	}
	


}

void tree::vrescale(float f)
{
	float m[4]={1,0,0,1};
	m[3]=f;
	float v[2]={0,0};
	affine_transform(m,v);
}

void tree::affine_transform(float m[4], float v[2])
{
	tnode *p=this;
	for(;;){
		fPoint q;
		q.x=m[0]*p->pos.x+m[1]*p->pos.y+v[0];
		q.y=m[2]*p->pos.x+m[3]*p->pos.y+v[1];
		p->pos=q;
		while(p->next && !p->prune){
			p=p->next;
			q.x=m[0]*p->pos.x+m[1]*p->pos.y+v[0];
			q.y=m[2]*p->pos.x+m[3]*p->pos.y+v[1];
			p->pos=q;
		}
		while(p->down==0){
			p=p->prev;
			if(p==0) return;
		}
		p=p->down;
	}

}





void tree::calcpos()
{

	calcdepth_sichtbar();
	bool fits=true;
	do{
		fits=true;
		calc_treewidth();
		calc_mindimensions();
		if(dsc.min_treewidth>dsc.treewidth){
			cout <<"Warning: Tree is " <<(dsc.min_treewidth-dsc.treewidth)/millimeter << "mm too wide.\n";
			float nw=ceil((dsc.picwidth+dsc.min_treewidth-dsc.treewidth)/millimeter)*millimeter;
			cout << "-->  Using \\width{" << nw/millimeter << "}\n";
			dsc.picwidth=nw;
			fits=false;
		}
		float mh=dsc.picheight-dsc.Rand.top-dsc.Rand.bottom;
		if(dsc.min_treeheight>mh){
			cout <<"Warning: Tree is " <<(dsc.min_treeheight-mh)/millimeter << "mm too high.\n";
			float nw=ceil((dsc.picheight+dsc.min_treeheight-mh)/millimeter)*millimeter;
			cout << "-->  Using \\height{" << nw/millimeter << "}\n";
			dsc.picheight=nw;
			fits=false;
		}
	}while(!fits);	
	
	if(dsc.uselength){
		calchpos_len();
	} else {
		if(dsc.variablelength){
			calchpos();
		} else {
			calchpos_fixed();
		}
	}
	calcvpos();		

//maximalen Rundungsradius bestimmen	
	dsc.maxrund=maxrund();
}




bool tree::reroot(tnode *nd)
//fuegt neue root vor nd ein
{

	if(nd==0) return false;
	tnode *p=nd->prev;
	if(p==0) return false; //alte Wurzel	
	int tnc=this->nchildren();
	if(tnc!=2){
		cerr << "Error: Original root must have 2 children (instead of "<<tnc<<")\n";
		//Der Fall, dass root nur einen Folgeknoten hat, muss anders behandelt werden
		//(spaeter einmal)
		//Bei mehr als einem Folgeknnoten von root ist ein reroot nicht moeglich
		return false;
	}

//Auf Weg von nr bis root durchlaufrichtung aendern
	tnode *q=nd;
	
	list<tnode*> tnd;
	while(p){
		p->removechild(q);
		tnd.push_back(q);
		q=p;
		p=p->prev;
	}
	tnd.push_back(next);
	removechild(next);
	list<tnode*>::iterator it=tnd.begin();
	insertchild(*it);
	++it;
	insertchild(*it);
	p=(*it);
	for(++it; it!=tnd.end(); ++it){
		p->insertchild(*it);
		p=(*it);
	}
	
	fill_labels(lb.total_nodes());
	return true;
}









